# -*- coding: utf-8 -*-


class RadishError(Exception):
    """
        General radish specific error
    """
    pass


class LanguageNotSupportedError(RadishError):
    """
        Raised if the language could not be found.
    """
    def __init__(self, language):
        self.language = language
        super(LanguageNotSupportedError, self).__init__("Language {0} could not be found".format(language))


class FeatureFileNotFoundError(RadishError):
    """
        Raised if a given feature file does not exist
    """
    def __init__(self, featurefile):
        self.featurefile = featurefile
        super(FeatureFileNotFoundError, self).__init__("Feature file '{0}': No such file".format(featurefile))


class FeatureFileSyntaxError(RadishError, SyntaxError):
    """
        If a a syntax error occured in a feature file
    """
    pass


class StepRegexError(RadishError, SyntaxError):
    """
        Raised if the step regex cannot be compiled.
    """
    def __init__(self, regex, step_func_name, re_error):
        self.regex = regex
        self.step_func_name = step_func_name
        self.re_error = re_error
        super(StepRegexError, self).__init__("Cannot compile regex '{0}' from step '{1}': {2}".format(regex, step_func_name, re_error))


class StepPatternError(RadishError, SyntaxError):
    """
        Raised if the steps pattern cannot be compiled.
    """
    def __init__(self, pattern, step_func_name, error):
        self.pattern = pattern
        self.step_func_name = step_func_name
        self.error = error
        super(StepPatternError, self).__init__("Cannot compile pattern '{0}' of step '{1}': {2}".format(pattern, step_func_name, error))


class SameStepError(RadishError):
    """
        Raised if two step regex are exactly the same.
    """
    def __init__(self, regex, func1, func2):
        self.regex = regex
        self.func1 = func1
        self.func2 = func2
        super(SameStepError, self).__init__("Cannot register step {0} with regex '{1}' because it is already used by step {2}".format(func2.__name__, regex, func1.__name__))


class StepDefinitionNotFoundError(RadishError):
    """
        Raised if the Matcher cannot find a step from the feature file in the registered steps.
    """
    def __init__(self, step):
        self.step = step
        super(StepDefinitionNotFoundError, self).__init__("Cannot find step definition for step '{0}' in {1}:{2}".format(step.sentence, step.path, step.line))


class RunnerEarlyExit(RadishError):
    """
        Raised if the runner has to exit to run.
    """
    pass


class HookError(RadishError):
    """
        Raised if an exception was raised inside a hook
    """
    def __init__(self, hook_function, failure):
        self.hook_function = hook_function
        self.failure = failure
        super(HookError, self).__init__("Hook '{0}' from {1}:{2} raised: '{3}: {4}'".format(hook_function.__name__, hook_function.__code__.co_filename, hook_function.__code__.co_firstlineno, failure.name, failure.reason))


class ScenarioNotFoundError(RadishError):
    """
        Raised if a scenario cannot be found
    """
    def __init__(self, scenario_id, amount_of_scenarios):
        self.scenario_id = scenario_id
        self.amount_of_scenarios = amount_of_scenarios
        super(ScenarioNotFoundError, self).__init__("No scenario with id {0} found. Specify a scenario id between 1 and {1}".format(scenario_id, amount_of_scenarios))


class ValidationError(RadishError):
    """
        Raised by the user if a step is somehow not valid
    """
    pass


class FeatureTagNotFoundError(RadishError):
    """
        Raised if a feature tag cannot be found
    """
    def __init__(self, tag):
        self.tag = tag
        super(FeatureTagNotFoundError, self).__init__("No features found with tag '{0}'".format(tag))


class ScenarioTagNotFoundError(RadishError):
    """
        Raised if a scenario tag cannot be found
    """
    def __init__(self, tag):
        self.tag = tag
        super(ScenarioTagNotFoundError, self).__init__("No scenarios found with tag '{0}'".format(tag))



# -*- coding: utf-8 -*-

"""
    This module provides a class to represent a Scenario
"""

from .model import Model
from .stepmodel import Step


class Scenario(Model):
    """
        Represents a Scenario
    """
    def __init__(self, id, keyword, sentence, path, line, parent, tags=None, preconditions=None):
        super(Scenario, self).__init__(id, keyword, sentence, path, line, parent, tags)
        self.absolute_id = None
        self.preconditions = preconditions or []
        self.steps = []
        self.context = self.Context()

    @property
    def state(self):
        """
            Returns the state of the scenario
        """
        for step in self.steps:
            if step.state is not Step.State.PASSED:
                return step.state
        return Step.State.PASSED

    @property
    def constants(self):
        """
            Returns all constants
        """
        constants = self.context.constants
        for name, value in constants:
            for parent_name, parent_value in self.parent.constants:
                value = value.replace("${{{0}}}".format(parent_name), parent_value)
        constants.extend(self.parent.constants)
        return constants

    @property
    def all_steps(self):
        """
            Returns all steps from all preconditions in the correct order
        """
        steps = []
        for precondition in self.preconditions:
            steps.extend(precondition.all_steps)
        steps.extend(self.steps)
        return steps

    @property
    def failed_step(self):
        """
            Returns the first failed step
        """
        for step in self.steps:
            if step.state == Step.State.FAILED:
                return step
        return None

    def has_to_run(self, scenario_choice, feature_tags, scenario_tags):
        """
            Returns wheiter the scenario has to run or not

            :param list scenario_choice: the scenarios to run. If None all will run
        """
        if not scenario_choice and not feature_tags and not scenario_tags:
            return True

        in_choice = self.absolute_id in (scenario_choice or [])

        in_tags = False
        if scenario_tags:
            in_tags = any(t for t in self.tags if t.name in scenario_tags)

        feature_has_to_run = False
        if feature_tags:
            feature_has_to_run = any(t for t in self.parent.tags if t.name in feature_tags)

        return in_choice or in_tags or feature_has_to_run

    def after_parse(self):
        """
            This method is called after the scenario is completely parsed.
            Actions to do:
                * number steps
                * fix parent of precondition steps
        """
        for step_id, step in enumerate(self.all_steps, start=1):
            step.id = step_id
            if step.parent != self:
                step.as_precondition = step.parent
                step.parent = self



# -*- coding: utf-8 -*-

"""
    Feature file parser.
    One Feature file parser instance is able to parse one feature file.
"""

import os
import codecs
import re
import json

from .exceptions import RadishError, FeatureFileSyntaxError, LanguageNotSupportedError
from .feature import Feature
from .scenario import Scenario
from .scenariooutline import ScenarioOutline
from .scenarioloop import ScenarioLoop
from .stepmodel import Step


class Keywords(object):
    """
        Represent config object for gherkin keywords.
    """
    def __init__(self, feature, scenario, scenario_outline, examples, scenario_loop, iterations):
        self.feature = feature
        self.scenario = scenario
        self.scenario_outline = scenario_outline
        self.examples = examples
        self.scenario_loop = scenario_loop
        self.iterations = iterations


class FeatureParser(object):
    """
        Class to parse a feature file.
        A feature file contains just one feature.
    """

    LANGUAGE_LOCATION = os.path.join(os.path.dirname(__file__), "languages")
    DEFAULT_LANGUAGE = "en"

    class State(object):
        """
            Represents the parser state
        """
        INIT = "init"
        FEATURE = "feature"
        SCENARIO = "scenario"
        SCENARIO_OUTLINE = "scenario_outline"
        STEP = "step"
        EXAMPLES = "examples"
        EXAMPLES_ROW = "examples_row"
        SCENARIO_LOOP = "scenario_loop"
        STEP_TEXT = "step_text"

    def __init__(self, core, featurefile, featureid, language="en"):
        if not os.path.exists(featurefile):
            raise OSError("Feature file at '{0}' does not exist".format(featurefile))

        self._core = core
        self._featureid = featureid
        self._featurefile = featurefile
        self.keywords = {}
        self._keywords_delimiter = ":"

        self._current_state = FeatureParser.State.FEATURE
        self._current_line = 0
        self._current_tags = []
        self._current_preconditions = []
        self._current_constants = []
        self._in_step_text = False
        self.feature = None

        self._load_language(language)

    def _load_language(self, language=None):
        """
            Loads all keywords of the given language

            :param string language: the lanugage to use for the feature files.
                                    if None is given `radish` tries to detect the language.

            :returns: if the language could be loaded or not
            :rtype: bool

            :raises LanguageNotSupportedError: if the given language is not supported by radish
        """
        if not language:  # try to detect language
            raise NotImplementedError("Auto detect language is not implemented yet")

        language_path = os.path.join(self.LANGUAGE_LOCATION, language + ".json")
        try:
            with codecs.open(language_path, "rb", "utf-8") as f:
                language_pkg = json.load(f)
        except IOError:
            raise LanguageNotSupportedError(language)

        self.keywords = Keywords(**language_pkg["keywords"])

    def parse(self):
        """
            Parses the feature file of this `FeatureParser` instance

            :returns: if the parsing was successful or not
            :rtype: bool
        """
        with codecs.open(self._featurefile, "rb", "utf-8") as f:
            for line in f.readlines():
                self._current_line += 1
                line = line.strip()
                if not line:  # line is empty
                    continue

                if line.startswith("#"):
                    # try to detect feature file language
                    language = self._detect_language(line)
                    if language:
                        self._load_language(language)

                    continue

                if self.feature and self._detect_feature(line):
                    raise FeatureFileSyntaxError("radish supports only one Feature per feature file")

                if not self._parse_context(line):
                    raise FeatureFileSyntaxError("Syntax error in feature file {0} on line {1}".format(self._featurefile, self._current_line))
        if not self.feature:
            raise FeatureFileSyntaxError("No Feature found in file {0}".format(self._featurefile))

        if self.feature.scenarios:
            self.feature.scenarios[-1].after_parse()

    def _parse_context(self, line):
        """
            Parses arbitrary context from a line

            :param string line: the line to parse from
        """
        parse_context_func = getattr(self, "_parse_" + self._current_state)
        if not parse_context_func:
            raise RadishError("FeatureParser state {0} is not support".format(self._current_state))

        return parse_context_func(line)

    def _parse_feature(self, line):
        """
            Parses a Feature Sentence

            The `INIT` state is used as initiale state.

            :param string line: the line to parse from
        """
        detected_feature = self._detect_feature(line)
        if not detected_feature:
            tag = self._detect_tag(line)
            if tag:
                self._current_tags.append(Feature.Tag(tag[0], tag[1]))
                if tag[0] == "constant":
                    name, value = self._parse_constant(tag[1])
                    self._current_constants.append((name, value))
                return True

            return False

        self.feature = Feature(self._featureid, self.keywords.feature, detected_feature, self._featurefile, self._current_line, self._current_tags)
        self.feature.context.constants = self._current_constants
        self._current_state = FeatureParser.State.SCENARIO
        self._current_tags = []
        self._current_constants = []
        return True

    def _parse_scenario(self, line):
        """
            Parses a Feature context

            :param string line: the line to parse from
        """
        detected_scenario = self._detect_scenario(line)
        scenario_type = Scenario
        keywords = (self.keywords.scenario,)
        if not detected_scenario:
            detected_scenario = self._detect_scenario_outline(line)
            scenario_type = ScenarioOutline
            keywords = (self.keywords.scenario_outline, self.keywords.examples)

            if not detected_scenario:
                detected_scenario = self._detect_scenario_loop(line)
                if not detected_scenario:
                    tag = self._detect_tag(line)
                    if tag:
                        self._current_tags.append(Scenario.Tag(tag[0], tag[1]))
                        if tag[0] == "precondition":
                            scenario = self._parse_precondition(tag[1])
                            self._current_preconditions.append(scenario)
                        elif tag[0] == "constant":
                            name, value = self._parse_constant(tag[1])
                            self._current_constants.append((name, value))
                        return True

                    self.feature.description.append(line)
                    return True

                detected_scenario, iterations = detected_scenario  # pylint: disable=unpacking-non-sequence
                scenario_type = ScenarioLoop
                keywords = (self.keywords.scenario_loop, self.keywords.iterations)

        if detected_scenario in self.feature:
            raise FeatureFileSyntaxError("Scenario with name '{0}' defined twice in feature '{1}'".format(detected_scenario, self.feature.path))

        scenario_id = 1
        if self.feature.scenarios:
            previous_scenario = self.feature.scenarios[-1]
            if hasattr(previous_scenario, "scenarios") and previous_scenario.scenarios:
                scenario_id = previous_scenario.scenarios[-1].id + 1
            else:
                scenario_id = previous_scenario.id + 1

        self.feature.scenarios.append(scenario_type(scenario_id, *keywords, sentence=detected_scenario, path=self._featurefile, line=self._current_line, parent=self.feature, tags=self._current_tags, preconditions=self._current_preconditions))
        self.feature.scenarios[-1].context.constants = self._current_constants
        self._current_tags = []
        self._current_preconditions = []
        self._current_constants = []

        if scenario_type == ScenarioLoop:
            self.feature.scenarios[-1].iterations = iterations
        self._current_state = FeatureParser.State.STEP
        return True

    def _parse_examples(self, line):
        """
            Parses the Examples header line

            :param string line: the line to parse from
        """
        if not isinstance(self.feature.scenarios[-1], ScenarioOutline):
            raise FeatureFileSyntaxError("Scenario does not support Examples. Use 'Scenario Outline'")

        self.feature.scenarios[-1].examples_header = [x.strip() for x in line.split("|")[1:-1]]
        self._current_state = FeatureParser.State.EXAMPLES_ROW
        return True

    def _parse_examples_row(self, line):
        """
            Parses an Examples row

            :param string line: the line to parse from
        """
        # detect next keyword
        if self._detect_scenario(line) or self._detect_scenario_outline(line) or self._detect_scenario_loop(line):
            self.feature.scenarios[-1].after_parse()
            return self._parse_scenario(line)

        example = ScenarioOutline.Example([x.strip() for x in line.split("|")[1:-1]], self._featurefile, self._current_line)
        self.feature.scenarios[-1].examples.append(example)
        return True

    def _parse_step(self, line):
        """
            Parses a single step

            :param string line: the line to parse from
        """
        # detect next keyword
        if self._detect_scenario(line) or self._detect_scenario_outline(line) or self._detect_scenario_loop(line) or self._detect_tag(line):
            self.feature.scenarios[-1].after_parse()
            return self._parse_scenario(line)

        if self._detect_step_text(line):
            self._current_state = self.State.STEP_TEXT
            return self._parse_step_text(line)

        if self._detect_table(line):
            self._parse_table(line)
            return True

        if self._detect_examples(line):
            self._current_state = FeatureParser.State.EXAMPLES
            return True

        step_id = len(self.feature.scenarios[-1].all_steps) + 1
        not_runable = isinstance(self.feature.scenarios[-1], (ScenarioOutline, ScenarioLoop))
        step = Step(step_id, line, self._featurefile, self._current_line, self.feature.scenarios[-1], not not_runable)
        self.feature.scenarios[-1].steps.append(step)
        return True

    def _parse_table(self, line):
        """
            Parses a step table row

            :param string line: the line to parse from
        """
        if not self.feature.scenarios[-1].steps:
            raise FeatureFileSyntaxError("Found step table without previous step definition on line {0}".format(self._current_line))

        self.feature.scenarios[-1].steps[-1].table.append([x.strip() for x in line.split("|")[1:-1]])
        return True

    def _parse_step_text(self, line):
        """
            Parses additional step text

            :param str line: the line to parse
        """
        if line.startswith('"""') and not self._in_step_text:
            self._in_step_text = True
            line = line[3:]

        if line.endswith('"""') and self._in_step_text:
            self._current_state = self.State.STEP
            self._in_step_text = False
            line = line[:-3]

        if line:
            self.feature.scenarios[-1].steps[-1].raw_text.append(line.strip())
        return True

    def _parse_precondition(self, arguments):
        """
            Parses scenario preconditions

            The arguments must be in format:
                File.feature: Some scenario

            :param str arguments: the raw arguments
        """
        match = re.search(r"(.*?\.feature): (.*)", arguments)
        if not match:
            raise FeatureFileSyntaxError("Scenario @precondition tag must have argument in format: 'test.feature: Some scenario'")

        feature_file_name, scenario_sentence = match.groups()
        feature_file = os.path.join(os.path.dirname(self._featurefile), feature_file_name)

        try:
            feature = self._core.parse_feature(feature_file)
        except RuntimeError as e:
            if str(e) == "maximum recursion depth exceeded":  # precondition cycling
                raise FeatureFileSyntaxError("Your feature '{0}' has cycling preconditions with '{1}: {2}' starting at line {3}".format(self._featurefile, feature_file_name, scenario_sentence, self._current_line))
            raise

        if scenario_sentence not in feature:
            raise FeatureFileSyntaxError("Cannot import precondition scenario '{0}' from feature '{1}': No such scenario".format(scenario_sentence, feature_file))

        return feature[scenario_sentence]

    def _parse_constant(self, arguments):
        """
            Parses tag arguments as a constant containing name and value

            The arguments must be in format:
                ConstantName: SomeValue
                ConstantName: 5

            :param str arguments: the raw arguments to parse
        """
        name, value = arguments.split(":", 1)
        return name.strip(), value.strip()

    def _detect_feature(self, line):
        """
            Detects a feature on the given line

            :param string line: the line to detect a feature

            :returns: if a feature was found on the given line
            :rtype: bool
        """
        if line.startswith(self.keywords.feature + self._keywords_delimiter):
            return line[len(self.keywords.feature) + len(self._keywords_delimiter):].strip()

        return None

    def _detect_scenario(self, line):
        """
            Detects a scenario on the given line

            :param string line: the line to detect a scenario

            :returns: if a scenario was found on the given line
            :rtype: bool
        """
        if line.startswith(self.keywords.scenario + self._keywords_delimiter):
            return line[len(self.keywords.scenario) + len(self._keywords_delimiter):].strip()

        return None

    def _detect_scenario_outline(self, line):
        """
            Detects a scenario outline on the given line

            :param string line: the line to detect a scenario outline

            :returns: if a scenario outline was found on the given line
            :rtype: bool
        """
        if line.startswith(self.keywords.scenario_outline + self._keywords_delimiter):
            return line[len(self.keywords.scenario_outline) + len(self._keywords_delimiter):].strip()

        return None

    def _detect_examples(self, line):
        """
            Detects an Examples block on the given line

            :param string line: the line to detect the Examples

            :returns: if an Examples block was found on the given line
            :rtype: bool
        """
        if line.startswith(self.keywords.examples + self._keywords_delimiter):
            return True

        return None

    def _detect_scenario_loop(self, line):
        """
            Detects a scenario loop on the given line

            :param string line: the line to detect a scenario loop

            :returns: if a scenario loop was found on the given line
            :rtype: string
        """
        match = re.search(r"^{0} (\d+):(.*)".format(self.keywords.scenario_loop), line)
        if match:
            return match.group(2).strip(), int(match.group(1))

        return None

    def _detect_table(self, line):
        """
            Detects a step table row on the given line

            :param string line: the line to detect the table row

            :returns: if an step table row was found or not
            :rtype: bool
        """
        return line.startswith("|")

    def _detect_step_text(self, line):
        """
            Detects the beginning of an additional step text block

            :param str line: the line to detect the step text block

            :returns: if a step text block was found or not
            :rtype: bool
        """
        return line.startswith('"""')

    def _detect_language(self, line):
        """
            Detects a language on the given line

            :param string line: the line to detect the language

            :returns: the language or None
            :rtype: str or None
        """
        match = re.search("^# language: (.*)", line)
        if match:
            return match.group(1)

        return None

    def _detect_tag(self, line):
        """
            Detects a tag on the given line

            :param string line: the line to detect the tag

            :returns: the tag or None
            :rtype: str or None
        """
        match = re.search(r"^@([^\s(]+)(?:\((.*?)\))?", line)
        if match:
            return match.group(1), match.group(2)

        return None



# -*- coding: utf-8 -*-

"""
    This module provides some functionality to diagnose thrown exceptions
"""


import sys
from functools import wraps
from colorful import colorful

from .terrain import world
from .exceptions import RadishError, FeatureFileSyntaxError, StepDefinitionNotFoundError, HookError, SameStepError


__RADISH_DOC__ = "https://github.com/radish-bdd/radish"


def write(text):
    """
        Writes the given text to the console
    """
    print(text)


def write_error(text):
    """
        Writes the given text to the console
    """
    write("{0}: {1}".format(colorful.bold_red("Error"), colorful.red(text)))


def write_failure(failure):
    """
        Writes the failure to the console
    """
    write("\n{0}".format(colorful.red(failure.traceback)))


def abort(return_code):
    """
        Aborts the program with the given return_code
    """
    sys.exit(return_code)


def error_oracle(func):
    """
        Decorator to diagnose thrown exceptions
    """
    @wraps(func)
    def _decorator(*args, **kwargs):
        """
            The actual decorator
        """
        try:
            return func(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            handle_exception(e)

    return _decorator


def catch_unhandled_exception(exc_type, exc_value, traceback):
    """
        Catch all unhandled exceptions
    """
    handle_exception(exc_value)


def handle_exception(exception):
    """
        Handle the given exception

        This will print more information about the given exception

        :param Exception exception: the exception to handle
    """
    if isinstance(exception, HookError):  # handle exception from hook
        write_error(exception)
        write_failure(exception.failure)
        abort(1)
    elif isinstance(exception, FeatureFileSyntaxError):
        write_error(exception)
        write("\nError Oracle says:")
        write("""You have a SyntaxError in your feature file!
Please have a look into the radish documentation to found out which
features radish supports and how you could use them:
Link: {0}
              """.format(__RADISH_DOC__))
        abort(1)
    elif isinstance(exception, StepDefinitionNotFoundError):
        write_error(exception)
        write("\nError Oracle says:")
        write("""There is no step defintion for '{0}'.
All steps should be declared in a module located in {1}.
For example you could do:

@step(r"{2}")
def my_step(step):
    raise NotImplementedError("This step is not implemented yet")
        """.format(exception.step.sentence, world.config.basedir, exception.step.sentence))
        abort(1)
    elif isinstance(exception, SameStepError):
        write_error(exception)
        write("\nError Oracle says:")
        write("""You have defined two step definitions with the same Regular Expression.
This is invalid since radish does not know which one is the one to go with.
If you have two similar step definition expressions but ones sentence is a subset of the other
you may want to add a $ to mark the sentence's end - take care of the code order - first comes first serves.
              """)
        abort(1)
    elif isinstance(exception, RadishError):
        write_error(exception)
        abort(1)
    elif isinstance(exception, KeyboardInterrupt):
        write("Aborted by the user...")
        abort(1)
    else:
        write_error(str(exception))
        abort(2)



# -*- coding: utf-8 -*-

import os
import sys
from docopt import docopt
from time import time

from . import __VERSION__
from .core import Core
from .loader import load_modules
from .matcher import Matcher
from .stepregistry import StepRegistry
from .hookregistry import HookRegistry
from .runner import Runner
from .extensionregistry import ExtensionRegistry
from .exceptions import FeatureFileNotFoundError, ScenarioNotFoundError, FeatureTagNotFoundError, ScenarioTagNotFoundError
from .errororacle import error_oracle, catch_unhandled_exception
from .terrain import world
from . import utils


def setup_config(arguments):
    """
        Parses the docopt arguments and creates a configuration object in terrain.world
    """
    world.config = lambda: None
    for key, value in arguments.items():
        config_key = key.replace("--", "").replace("-", "_").replace("<", "").replace(">", "")
        setattr(world.config, config_key, value)


def show_features(core):
    """
        Show the parsed features
    """
    # set needed configuration
    world.config.write_steps_once = True
    if not sys.stdout.isatty():
        world.config.no_ansi = True

    runner = Runner(HookRegistry(), dry_run=True)
    runner.start(core.features_to_run, marker="show")
    return 0


def run_features(core):
    """
        Run the parsed features

        :param Core core: the radish core object
    """
    # set needed configuration
    world.config.expand = True

    # load user's custom python files
    load_modules(world.config.basedir)

    # match feature file steps with user's step definitions
    Matcher.merge_steps(core.features, StepRegistry().steps)

    # run parsed features
    if world.config.marker == "time.time()":
        world.config.marker = int(time())

    # scenario choice
    amount_of_scenarios = sum(len(f.scenarios) for f in core.features_to_run)
    if world.config.scenarios:
        world.config.scenarios = [int(s) for s in world.config.scenarios.split(",")]
        for s in world.config.scenarios:
            if not (0 < s <= amount_of_scenarios):
                raise ScenarioNotFoundError(s, amount_of_scenarios)

    # tags
    if world.config.feature_tags:
        world.config.feature_tags = [t for t in world.config.feature_tags.split(",")]
        for tag in world.config.feature_tags:
            if not any(f for f in core.features if tag in [t.name for t in f.tags]):
                raise FeatureTagNotFoundError(tag)

    if world.config.scenario_tags:
        world.config.scenario_tags = [t for t in world.config.scenario_tags.split(",")]
        for tag in world.config.scenario_tags:
            if not any(s for f in core.features for s in f.scenarios if tag in [t.name for t in s.tags]):
                raise ScenarioTagNotFoundError(tag)

    runner = Runner(HookRegistry(), early_exit=world.config.early_exit)
    return runner.start(core.features_to_run, marker=world.config.marker)


@error_oracle
def main():
    """
Usage:
    radish show <features>
           [--expand]
           [--no-ansi]
    radish <features>...
           [-b=<basedir> | --basedir=<basedir>]
           [-e | --early-exit]
           [--debug-steps]
           [-t | --with-traceback]
           [-m=<marker> | --marker=<marker>]
           [-p=<profile> | --profile=<profile>]
           [-d | --dry-run]
           [-s=<scenarios> | --scenarios=<scenarios>]
           [--shuffle]
           [--feature-tags=<feature_tags>]
           [--scenario-tags=<scenario_tags>]
           {0}
    radish (-h | --help)
    radish (-v | --version)

Arguments:
    features                                    feature files to run

Options:
    -h --help                                   show this screen
    -v --version                                show version
    -e --early-exit                             stop the run after the first failed step
    --debug-steps                               debugs each step
    -t --with-traceback                         show the Exception traceback when a step fails
    -m=<marker> --marker=<marker>               specify the marker for this run [default: time.time()]
    -p=<profile> --profile=<profile>            specify the profile which can be used in the step/hook implementation
    -b=<basedir> --basedir=<basedir>            set base dir from where the step.py and terrain.py will be loaded [default: $PWD/radish]
    -d --dry-run                                make dry run for the given feature files
    -s=<scenarios> --scenarios=<scenarios>      only run the specified scenarios (comma separated list)
    --shuffle                                   shuttle run order of features and scenarios
    --feature-tags=<feature_tags>               only run features with the given tags
    --scenario-tags=<scenario_tags>             only run scenarios with the given tags
    --expand                                    expand the feature file (all preconditions)
    {1}

(C) Copyright 2013 by Timo Furrer <tuxtimo@gmail.com>
    """
    # load extensions
    load_modules(os.path.join(os.path.dirname(__file__), "extensions"))

    extensions = ExtensionRegistry()
    main.__doc__ = main.__doc__.format(extensions.get_options(), extensions.get_option_description())

    sys.excepthook = catch_unhandled_exception

    arguments = docopt("radish {0}\n{1}".format(__VERSION__, main.__doc__), version=__VERSION__)

    # store all arguments to configuration dict in terrain.world
    setup_config(arguments)

    # load needed extensions
    extensions.load(world.config)

    core = Core()

    feature_files = []
    for given_feature in world.config.features:
        if not os.path.exists(given_feature):
            raise FeatureFileNotFoundError(given_feature)

        if os.path.isdir(given_feature):
            feature_files.extend(utils.recursive_glob(given_feature, "*.feature"))
            continue

        feature_files.append(given_feature)

    core.parse_features(feature_files)

    if not core.features:
        print("Error: no features given")
        return 1

    argument_dispatcher = [
        ((lambda: world.config.show), show_features),
        ((lambda: True), run_features)
    ]

    # radish command dispatching
    for to_run, method in argument_dispatcher:
        if to_run():
            return method(core)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

"""
    Providing radish core functionality.
"""

from threading import Lock
from collections import OrderedDict

from .parser import FeatureParser


# FIXME: rename
class Core(object):
    """
        Provide some core functionalities like parsing and storing of the feature files
    """
    def __init__(self):
        self.features = []
        self._features_to_run = OrderedDict()
        self._feature_id_lock = Lock()
        self._feature_id = 0
        self._scenario_id_lock = Lock()
        self._scenario_id = 0

    @property
    def features_to_run(self):
        """
            Return all parsed features which are to run
        """
        return [f for f in self._features_to_run.values()]

    @property
    def next_feature_id(self):
        """
            Returns the next feature id
        """
        with self._feature_id_lock:
            self._feature_id += 1
            return self._feature_id

    @property
    def next_scenario_id(self):
        """
            Returns the next scenario id
        """
        with self._scenario_id_lock:
            self._scenario_id += 1
            return self._scenario_id

    def parse_features(self, feature_files):
        """
            Parses the given feature files
        """
        for featurefile in feature_files:
            feature = self.parse_feature(featurefile, self.next_feature_id)

            for scenario in feature.scenarios:
                scenario.absolute_id = self.next_scenario_id

            self._features_to_run[featurefile] = feature

    def parse_feature(self, featurefile, featureid=0):
        """
            Parses the given feature file
            If the feature is alreay parsed then it will just return it

            :returns: the parsed feature
            :rtype: Feature
        """
        featureparser = FeatureParser(self, featurefile, featureid)
        featureparser.parse()

        self.features.append(featureparser.feature)
        return featureparser.feature



# -*- coding: utf-8 -*-

"""
    This module provides a radish extension to make
    coverage measurements.
"""


from radish.extensionregistry import extension
from radish.hookregistry import before, after
from radish.terrain import world

import sys
from coverage import Coverage


@extension
class CodeCoverage(object):
    """
        Code Coverage radish extension
    """
    OPTIONS = [
        ("--with-coverage", "enable code coverage"),
        ("--cover-packages=<cover_packages>", "specify source code package")
    ]
    LOAD_IF = staticmethod(lambda config: config.with_coverage)
    LOAD_PRIORITY = 70

    def __init__(self):
        before.all(self.coverage_start)
        after.all(self.coverage_stop)

        if world.config.cover_packages:
            cover_packages = world.config.cover_packages.split(",")
        else:
            cover_packages = []
        self.coverage = Coverage(source=cover_packages)

    def coverage_start(self, features, marker):
        """
            Start the coverage measurement
        """
        self.coverage.load()
        self.coverage.start()

    def coverage_stop(self, features, marker):
        """
            Stop the coverage measurement
            and create report
        """
        self.coverage.stop()
        self.coverage.save()
        self.coverage.report(file=sys.stdout)



# -*- coding: utf-8 -*-

"""
    This module provides a class to match the feature file steps with the registered steps from the registry
"""

import re
import parse
from collections import namedtuple

from .argexpregistry import ArgExpRegistry
from .exceptions import StepDefinitionNotFoundError, StepPatternError

StepMatch = namedtuple("StepMatch", ["args", "kwargs", "func"])


class Matcher(object):
    """
        Matches steps from the feature files with the registered steps
    """
    @classmethod
    def merge_steps(cls, features, steps):
        """
            Merges steps from the given features with the given steps

            :param list features: the features
            :param dict steps: the steps
        """
        # FIXME: fix cycle-import ... Matcher -> ScenarioOutline -> Step -> Matcher
        from .scenariooutline import ScenarioOutline
        for feature in features:
            for scenario in feature.all_scenarios:
                if isinstance(scenario, ScenarioOutline):
                    continue  # ScenarioOutline steps do not have to be merged

                for step in scenario.steps:
                    cls.merge_step(step, steps)

    @classmethod
    def merge_step(cls, step, steps):
        """
            Merges a single step with the registered steps

            :param Step step: the step from a feature file to merge
            :param list steps: the registered steps
        """
        match = cls.match(step.expanded_sentence, steps)
        if not match or not match.func:
            raise StepDefinitionNotFoundError(step)

        step.definition_func = match.func
        step.arguments = match.args
        step.keyword_arguments = match.kwargs

    @staticmethod
    def match(sentence, steps):
        """
            Tries to find a match from the given sentence with the given steps

            :param string sentence: the step sentence to match
            :param dict steps: the available registered steps

            :returns: the arguments and the func which were matched
            :rtype: tuple
        """
        for pattern, func in steps.items():
            if isinstance(pattern, re._pattern_type):  # pylint: disable=protected-access
                match = pattern.search(sentence)
                if match:
                    return StepMatch(args=match.groups(), kwargs=match.groupdict(), func=func)
            else:
                try:
                    compiled = parse.compile(pattern, ArgExpRegistry().expressions)
                except ValueError as e:
                    raise StepPatternError(pattern, func.__name__, e)

                match = compiled.search(sentence)
                if match:
                    return StepMatch(args=match.fixed, kwargs=match.named, func=func)

        return None



# -*- coding: utf-8 -*-

import re
from tests.base import *

from radish.matcher import Matcher
from radish.feature import Feature
from radish.scenario import Scenario
from radish.stepmodel import Step
from radish.exceptions import StepDefinitionNotFoundError


class MatcherTestCase(RadishTestCase):
    """
        Tests the Matcher class
    """
    def test_match_steps(self):
        """
            Test matching steps from feature files with registered steps
        """
        matcher = Matcher()
        steps = {re.compile(r"Given I have the number (\d+)"): "some_func", re.compile(r"I add (\d+) to my number"): "some_other_func"}

        arguments, keyword_arguments, func = matcher.match("Given I have the number 5", steps)
        arguments.should.be.equal(("5",))
        keyword_arguments.should.be.equal({})
        func.should.be.equal("some_func")

        arguments, keyword_arguments, func = matcher.match("When I add 2 to my number", steps)
        arguments.should.be.equal(("2",))
        keyword_arguments.should.be.equal({})
        func.should.be.equal("some_other_func")

        match = matcher.match("when I call a non-existing step", steps)
        match.should.be.none

    def test_merge_steps(self):
        """
            Test merging steps from feature files with registered steps
        """
        matcher = Matcher()
        steps = {re.compile(r"Given I have the number (\d+)"): "some_func", re.compile(r"I add (\d+) to my number"): "some_other_func"}

        feature = Feature(1, "Feature", "Some feature", "test.feature", 1)
        scenario = Scenario(1, "Scenario", "Adding numbers", "test.feature", 2, feature)
        scenario.steps.append(Step(1, "Given I have the number 5", "test.feature", 3, scenario, False))
        scenario.steps.append(Step(2, "When I add 2 to my number", "test.feature", 4, scenario, False))
        feature.scenarios.append(scenario)

        matcher.merge_steps([feature], steps)

        scenario.steps[0].definition_func.should.be.equal("some_func")
        scenario.steps[0].arguments.should.be.equal(("5",))
        scenario.steps[1].definition_func.should.be.equal("some_other_func")
        scenario.steps[1].arguments.should.be.equal(("2",))

    def test_merge_non_existing_step(self):
        """
            Test merging non existing step
        """
        matcher = Matcher()
        steps = {re.compile(r"Given I have the number (\d+)"): "some_func", re.compile(r"I add (\d+) to my number"): "some_other_func"}

        feature = Feature(1, "Feature", "Some feature", "test.feature", 1)
        scenario = Scenario(1, "Scenario", "Adding numbers", "test.feature", 2, feature)
        scenario.steps.append(Step(1, "When I call a non-existing step", "test.feature", 3, scenario, False))
        feature.scenarios.append(scenario)

        matcher.merge_steps.when.called_with([feature], steps).should.throw(StepDefinitionNotFoundError, "Cannot find step definition for step 'When I call a non-existing step' in test.feature:3")
